/***********************************
 Data Visualization - 3D Project - "Global Migration Movements": DeweyClass Changes over Time
 
 Author: Ingmar Sturm
 Supervisor: George Legrady
 
 Purpose:  Visualize global migration movements over time.
 
 Usage:    1. Press the key 's' to take a screenshot
           2. Press the keys 0, 1, 2, 3 to switch between the years 
           3. Enter a 3-letter ISO code to filter by that year
           4. Hover over spheres to display 3-letter ISO code
           
 Created and tested on Processing version 4.02a
 
 ***********************************/

import peasy.*;
PeasyCam cam;

import controlP5.*;
ControlP5 cp5;

import java.util.Arrays; // for some arraymagic

Table populationsTable;
int numRowsPop, numColumnsPop;
Table migrationTable;
int numRowsMig, numColumnsMig;

ArrayList<PVector> list = new ArrayList();

int yearFlag = 0;

// create a new array of countries
ArrayList<Country> countries = new ArrayList<Country>();

//IntList population;

void setup() {
  size(1300, 1300, P3D);
  cam = new PeasyCam(this, 200);
  pixelDensity(displayDensity());

  cp5 = new ControlP5(this);
  cp5.setAutoDraw(false);

  PFont font = createFont("ArialCE.ttf", 10, true);

  Textfield t = cp5.addTextfield("Enter Country Code")
    .setPosition(10, height-65)
    .setSize(200, 40)
    .setFont(createFont("ArialCE.ttf", 15, true))
    .setFocus(true)
    .setColor(color(255, 255, 255))
    .setColorBackground(color(16, 102, 141))
    //.setColorActive(0xff00ff00)         //Green
    //.setColorBackground(0xff880000)  //Dark Red
    ;

  Label label = t.getCaptionLabel();
  label.setFont(createFont("ArialCE.ttf", 5, true));


  populationsTable = loadTable("population_averages.csv", "header");
  numRowsPop = populationsTable.getRowCount();
  println(numRowsPop);

  // read in the migration movements file
  migrationTable = loadTable("migration_flows.csv", "header");
  numRowsMig = migrationTable.getRowCount();

  // Set random position of spheres

  randomSeed(0);

  int yearFlag = 0;

  //population = new IntList();
  ////get populations for starting values
  //for (int i=0; i < numRowsPop; i ++) {
  //  population.append(populationsTable.getInt(i, 2));
  //}

  color from = color(204, 102, 0);
  color to = color(0, 102, 153);
  // define increments to advance the color gradient by this much:
  float fraction = 1.0/numRowsPop;
  float cumFraction = fraction;



  for (int i=0; i<numRowsPop; i++) {
    //String countryName = populationsTable.getString(i, 0);

    int population = populationsTable.getInt(i, 2);
    float posx=random(0, 600);
    float posy=random(0, 600);
    float posz=random(0, 600);
    color sphereColor =  lerpColor(from, to, cumFraction);
    String origin = "";
    ArrayList<String> destination = new ArrayList<String>();
    ArrayList<Integer> mig1990 = new ArrayList<Integer>();
    ArrayList<Integer> mig1995 = new ArrayList<Integer>();
    ArrayList<Integer> mig2000 = new ArrayList<Integer>();
    ArrayList<Integer> mig2005 = new ArrayList<Integer>();
    for (int j=0; j < numRowsMig; j++) {
      //println(migrationTable.getString(j, 1));
      if (populationsTable.getString(i, 1).equals(migrationTable.getString(j, 1))) {
        origin = migrationTable.getString(j, 1);
        destination.add(migrationTable.getString(j, 3));
        mig1990.add(migrationTable.getInt(j, 4));
        mig1995.add(migrationTable.getInt(j, 5));
        mig2000.add(migrationTable.getInt(j, 6));
        mig2005.add(migrationTable.getInt(j, 7));
      }
    }
    //println(origin, population, posx, posy, posz, sphereColor, destination, mig1990, mig1995, mig2000, mig2005, "\n");
    countries.add(new Country(origin, population, posx, posy, posz, sphereColor, destination, mig1990, mig1995, mig2000, mig2005));
    cumFraction = cumFraction + fraction;
  }
}

void draw() {
  background(0);

  // Draw red sphere at origin for easier navigation
  //fill(178, 34, 34);
  //sphere(10);

  //DRAW SPHERES //

  // Draw spheres by color
  for (int i = 0; i < countries.size(); i++) {
    // map the initial population data to -50 to 50
    float population_scaled = log(map(countries.get(i).population, 9124, 1321040833, 100, 10000));

    pushMatrix();
    // define cumulative fraction to feed to lerpColor
    noStroke();
    // fill the sphere with a given color
    fill(countries.get(i).sphereColor);
    lights();
    translate2(countries.get(i).coordinates); //translate2 is a user-defined function that takes a pVector as argument
    sphere(population_scaled);

    float mouseObjectDistance = sq(mouseX-screenX(0, 0, 0))+sq(mouseY-screenY(0, 0, 0));
    if (mouseObjectDistance < 100)
    {
      float[] rotations = cam.getRotations();
      rotateX(rotations[0]);
      rotateY(rotations[1]);
      rotateZ(rotations[2]);
      stroke(0, 0, 0, 0);
      fill(255, 255, 255);
      textMode(SHAPE);
      textSize(30);
      text(countries.get(i).origin, 0, 0);
    }
    popMatrix();
  }

  // END DRAW SPHERES //

  // DRAW VERTICES //

  //Conditional on textfield entry, either show all lines or only those of a specific country
  String countryCodeEntered = cp5.get(Textfield.class, "Enter Country Code").getText().toUpperCase();
  StringList originList =new StringList();
  for (int i = 0; i < countries.size(); i++) {
    originList.append(countries.get(i).origin);
  }
  int countryCodeIndex = originList.index(countryCodeEntered);
  if (!originList.hasValue(countryCodeEntered)) {
    noFill();
    beginShape();
    for (int i = 0; i < countries.size(); i++) {
      //float firstCoordinates = 0;
      //float secondCoordinates = 0;
      //println(countries.get(i).flowNumArray.get(yearFlag));
      for (int j = 0; j <  countries.get(i).flowNumArray.get(yearFlag).size(); j++) {
        //String currentOrigin = countries.get(i).origin;
        String currentDestination = countries.get(i).destination.get(j);
        PVector coordinatesOrigin = countries.get(i).coordinates;
        PVector coordinatesDestination = new PVector(0, 0, 0);
        for (int k =0; k<countries.size(); k++) {
          if (currentDestination.equals(countries.get(k).origin)) {
            coordinatesDestination = countries.get(k).coordinates;
            float vecDist = abs(coordinatesDestination.mag()-coordinatesOrigin.mag());
            PVector controlPoint1 = new PVector(((countries.get(i).posx + countries.get(k).posx)/2)+0.0, ((countries.get(i).posy + countries.get(k).posy)/2)+0.0, ((countries.get(i).posz + countries.get(k).posz)/2)+0.0);
            //PVector controlPoint2 = new PVector(((countries.get(i).posx + countries.get(k).posx)/2)+0.0, ((countries.get(i).posy + countries.get(k).posy)/2)+vecDist*0.5, ((countries.get(i).posz + countries.get(k).posz)/2)+0.0);
            stroke(countries.get(i).sphereColor);
            curveTightness(1.0/vecDist);
            float propStrokeWeight = map(float(countries.get(i).flowNumArray.get(yearFlag).get(j)), 0, 2677763, 0, 30);
            strokeWeight(propStrokeWeight);
            curveVertex2(coordinatesOrigin);
            //vertex2(PVector.add(coordinatesOrigin, new PVector(0, 5, 0)));
            stroke(countries.get(k).sphereColor);
            //curveVertex2(controlPoint1);
            curveVertex2(coordinatesDestination);
            //vertex2(PVector.add(coordinatesDestination, new PVector(0, 5, 0)));
            //vertex2(coordinatesDestination);
          }
        }
      }
    }
    endShape();
  } else {
    noFill();
    beginShape();
    for (int j = 0; j <  countries.get(countryCodeIndex).flowNumArray.get(yearFlag).size(); j++) {
      //String currentOrigin = countries.get(i).origin;
      String currentDestination = countries.get(countryCodeIndex).destination.get(j);
      PVector coordinatesOrigin = countries.get(countryCodeIndex).coordinates;
      PVector coordinatesDestination = new PVector(0, 0, 0);
      for (int k =0; k<countries.size(); k++) {
        if (currentDestination.equals(countries.get(k).origin)) {
          coordinatesDestination = countries.get(k).coordinates;
          float vecDist = abs(coordinatesDestination.mag()-coordinatesOrigin.mag());
          PVector controlPoint1 = new PVector(((countries.get(countryCodeIndex).posx + countries.get(k).posx)/2)+0.0, ((countries.get(countryCodeIndex).posy + countries.get(k).posy)/2)+0.0, ((countries.get(countryCodeIndex).posz + countries.get(k).posz)/2)+0.0);
          //PVector controlPoint2 = new PVector(((countries.get(i).posx + countries.get(k).posx)/2)+0.0, ((countries.get(i).posy + countries.get(k).posy)/2)+vecDist*0.5, ((countries.get(i).posz + countries.get(k).posz)/2)+0.0);
          stroke(countries.get(countryCodeIndex).sphereColor);
          curveTightness(1.0/vecDist);
          float propStrokeWeight = map(float(countries.get(countryCodeIndex).flowNumArray.get(yearFlag).get(j)), 0, 2677763, 0, 30);
          strokeWeight(propStrokeWeight);
          curveVertex2(coordinatesOrigin);
          //vertex2(PVector.add(coordinatesOrigin, new PVector(0, 5, 0)));
          stroke(countries.get(k).sphereColor);
          //curveVertex2(controlPoint1);
          curveVertex2(coordinatesDestination);
          //vertex2(PVector.add(coordinatesDestination, new PVector(0, 5, 0)));
          //vertex2(coordinatesDestination);
        }
      }
    }
    endShape();
  }

  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
}
